/**
 * @file purc-features.h
 * @author Vincent Wei (https://github.com/VincentWei)
 * @date 2021/07/05
 * @brief The API to test features enabled in compile time.
 *
 * Copyright (C) 2021 FMSoft <https://www.fmsoft.cn>
 *
 * This file is a part of PurC (short for Purring Cat), an HVML interpreter.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef PURC_PURC_FEATURES_H
#define PURC_PURC_FEATURES_H

#include <stdbool.h>
#include <stddef.h>

#include "purc-macros.h"

/**
 * purc_feature:
 *
 * Enum values used for testing features.
 *
 * @PURC_FEATURE_SOCKET_STREAM:
 *      The RWStream on Unix file descriptor or Windows socket.
 * @PURC_FEATURE_DOCTYPE_HTML:
 *      The HTML parser.
 * @PURC_FEATURE_DOCTYPE_XGML:
 *      The XGML parser.
 * @PURC_FEATURE_DOCTYPE_XML:
 *      The XML parser.
 * @PURC_FEATURE_REMOTE_FETCHER,
 *      The support for remote fetcher.
 * @PURC_FEATURE_RDRCM_THREAD,
 *      The support for thread-based renderer.
 * @PURC_FEATURE_RDRCM_SOCKET,
 *      The support for PURCMC-based renderer.
 * @PURC_FEATURE_RDRCM_HBDBUS,
 *      The support for HBDBus-based renderer.
 * @PURC_FEATURE_STREAM_HBDBUS:
 *      The listener on HBDBus.
 * @PURC_FEATURE_STREAM_WEB_SOCKET:
 *      The listener on WebSocket.
 * @PURC_FEATURE_STREAM_MQTT:
 *      The listener on MQTT.
 * @PURC_FEATURE_SSL:
 *      The support for secure socket connection.
 * @PURC_FEATURE_APP_AUTH:
 *      The support for app authentication.
 */
enum purc_feature {
    PURC_FEATURE_FIRST = 0,

    PURC_FEATURE_SOCKET_STREAM = PURC_FEATURE_FIRST,
    PURC_FEATURE_DOCTYPE_HTML,
    PURC_FEATURE_DOCTYPE_XML,
    PURC_FEATURE_DOCTYPE_XGML,
    PURC_FEATURE_REMOTE_FETCHER,
    PURC_FEATURE_RDRCM_THREAD,
    PURC_FEATURE_RDRCM_SOCKET,
    PURC_FEATURE_RDRCM_HBDBUS,
    PURC_FEATURE_STREAM_HBDBUS,
    PURC_FEATURE_STREAM_MQTT,
    PURC_FEATURE_STREAM_WEB_SOCKET,
    PURC_FEATURE_SSL,
    PURC_FEATURE_APP_AUTH,
    PURC_FEATURE_DNSSD,

    /* XXX: change this if you append a new feature. */
    PURC_FEATURE_LAST = PURC_FEATURE_DNSSD,
};

#define PURC_FEATURE_NR (PURC_FEATURE_LAST - PURC_FEATURE_FIRST + 1)

#define _PURC_ON      1
#define _PURC_OFF     0

#define PCA_ENABLE_SOCKET_STREAM        _PURC_ON
#define PCA_ENABLE_DOCTYPE_HTML         _PURC_ON
#define PCA_ENABLE_DOCTYPE_XML          _PURC_ON
#define PCA_ENABLE_DOCTYPE_XGML         _PURC_ON
#define PCA_ENABLE_REMOTE_FETCHER       _PURC_OFF
#define PCA_ENABLE_RDRCM_THREAD         _PURC_ON
#define PCA_ENABLE_RDRCM_SOCKET         _PURC_ON
#define PCA_ENABLE_RDRCM_HBDBUS         _PURC_ON
#define PCA_ENABLE_STREAM_HBDBUS        _PURC_ON
#define PCA_ENABLE_STREAM_MQTT          _PURC_ON
#define PCA_ENABLE_STREAM_WEB_SOCKET    _PURC_ON
#define PCA_ENABLE_SSL                  _PURC_ON
#define PCA_ENABLE_APP_AUTH             _PURC_OFF
#define PCA_ENABLE_DNSSD                _PURC_OFF

PCA_EXTERN_C_BEGIN

/**
 * purc_is_feature_enabled:
 *
 * @feature: the feature identifier.
 *
 * @see_also #purc_feature
 *
 * Returns: @true for enabled, otherwise false.
 */
PCA_EXPORT bool
purc_is_feature_enabled(enum purc_feature feature);

PCA_EXTERN_C_END

#endif /* not defined PURC_PURC_FEATURES_H */

